import numpy as np
from Terminatetensorflow.dense import *
from Terminatetensorflow.loss import *
from Terminatetensorflow.metrics import *
from Terminatetensorflow.LiveAnimation import *
from PIL import Image , ImageOps
  


class Model:
    def __init__(self, layers_dims, activation_fns):
        self.layers = []
        self.num_of_layers = len(layers_dims)

        for i in range(1, self.num_of_layers):
            self.layers.append(layer(layers_dims[i], layers_dims[i-1], activation_fns[i-1]))
        
        self.loss = None
    
    def predict(self, X):
        output = X
        for layer in self.layers:
            output = layer.forward_prop(output)

        _pred = np.array(output[0])        
        pred = np.zeros((10,X.shape[1]))  
        pred = np.argmax(output,axis=0)
        
        return np.atleast_2d(pred)
        
    def predict_ext_image(self, path):
        
        px = Image.open(path)     
        gray_image = ImageOps.grayscale(px)    
        new_image = gray_image.resize((28, 28))   
        
        pix_val = np.array(list(new_image.getdata()))    
        pix_val =np.atleast_2d(pix_val)
        pix_val = pix_val/255
        pred = Model.predict(self,pix_val.T)
        print("Predicted val", pred)
    
    def fit(self, X, y, learning_rate, metrics, epochs, batch_size, costt, optimizer):
        y_class = np.zeros((10,y.shape[1]))  
        y_class = np.atleast_2d(np.argmax(y,axis=0))
        batch_number = int(-(-X.shape[1]/batch_size // 1))
        metrics_output = {}
        
        for i in range(epochs):
            file_name = "animation_loss.txt"
            file_name_2 = "animation_acc.txt"
            file = open(file_name,"a+")
            file_2 = open(file_name_2,"a+")
            
            for j in range(batch_number):
                if j == batch_number:
                        
                    output = X[:,batch_size*(j):]
                    A = []
                    A.append(output)
                    for layer in self.layers:
                        output = layer.forward_prop(output)
                        A.append(output)                
                    
                    if (costt == "multiclass"):
                        self.loss = compute_multiclass_loss(y[:,batch_size*(j):], output)
                    elif (costt == "mse"):
                        self.loss = compute_mse_loss(y[:,batch_size*(j):], output)
                        
                    output_layer = True
                    dA_prev = 0
                    k = self.num_of_layers - 1
                    for layer in reversed(self.layers):
                        dA_prev,_,_ = layer.backward_prop(A[k - 1], dA_prev, y[:,batch_size*(j):], output_layer)
                        output_layer = False
                        k = k - 1
                    
                    if optimizer == "momentum":
                        for layer in self.layers:
                            layer.update_parameters_momentum(learning_rate)
                    elif optimizer == "rmsprop":
                        for layer in self.layers:
                            layer.update_parameters_rmsprop(learning_rate)
                    elif optimizer == "adam":
                        for layer in self.layers:
                            layer.update_parameters_adam(learning_rate)
                    else:
                        for layer in self.layers:
                            layer.update_parameters(learning_rate)
                         
                         
                else:
                    
                    output = X[:,batch_size*(j):batch_size*(j+1)]
                    A = []
                    A.append(output)
                    for layer in self.layers:
                        output = layer.forward_prop(output)
                        A.append(output)                
                    
                    if (costt == "multiclass"):
                        self.loss = compute_multiclass_loss(y[:,batch_size*(j):batch_size*(j+1)], output)
                    elif (costt == "mse"):
                        self.loss = compute_mse_loss(y[:,batch_size*(j):batch_size*(j+1)], output)
                        
                    output_layer = True
                    dA_prev = 0
                    k = self.num_of_layers - 1
                    for layer in reversed(self.layers):
                        dA_prev,_,_ = layer.backward_prop(A[k - 1], dA_prev, y[:,batch_size*(j):batch_size*(j+1)], output_layer)
                        output_layer = False
                        k = k - 1
                    
                    if optimizer == "rmsprop":
                        for layer in self.layers:
                            layer.update_parameters_rmsprop(learning_rate)
                    elif optimizer == "adam":
                        for layer in self.layers:
                            layer.update_parameters_adam(learning_rate)
                    elif optimizer == "momentum":
                        for layer in self.layers:
                            layer.update_parameters_momentum(learning_rate)
                    else:
                        for layer in self.layers:
                            layer.update_parameters(learning_rate)
                        
            print("Epoch", i, "->")
            print("\t\tLoss =", ("%.4f" % self.loss ))
            metrics_output['loss'+str(i)] = self.loss
            pred = self.predict(X)
        
            file.write(f'{i},{self.loss}'+'\n')
            file.close()
            

            for j in range(len(metrics)-1):
                Type = metrics[len(metrics)-1]
                if metrics[j] == "accuracy":
                    acc = accuracy_metrics(pred.T, y_class.T) * 100
                    metrics_output["accuracy"+str(i)] = acc
                    print("\t\tAccuracy =", ("%.4f" % acc ))
                elif metrics[j] == "f1 score":
                    f1 = F1_score(pred.T, y_class.T,Type)
                    metrics_output["f1"+str(i)] = f1
                    print("\t\tF1 score =", ("%.4f" % f1 ))
                elif metrics[j] == "precision":
                    P = Precision(pred.T,y_class.T,Type)
                    metrics_output["precision"+str(i)] = P
                    print("\t\tPrecision =",("%.4f" % P))
                elif metrics[j] == "recall":
                    R = Recall(pred.T,y_class.T,Type)
                    metrics_output["recall"+str(i)] = R
                    print("\t\tRecall =",("%.4f" % R))                    
                elif metrics[j] == "confusion matrix":
                    np.set_printoptions(suppress=True)
                    conf_mat = confusion_matrix(pred.T,y_class.T)
                    print("\t\tConfusion Matrix : \n",conf_mat)
                    
                file_2.write(f'{i},{acc}'+'\n')
                file_2.close()
                
                
        # Loss GIF         
        draw_ = draw(epochs,file_name)
        draw_.save_gif()
        
        # ACC GIF
        draw_2 = draw(epochs,file_name_2)
        draw_2.save_gif()
        
        draw_.terminate()
        draw_2.terminate()

        print("\n___________________________________________________\n")
        print("Final cost:", ("%.5f" % self.loss ))
        
        for j in range(len(metrics)):
                if metrics[j] == "accuracy":
                    print("Final training accuracy:", ("%.5f" % acc ), "%")
                elif metrics[j] == "f1 score":
                     print("Final training value for F1 score:", ("%.5f" % f1 ))
                elif metrics[j] == "precision":
                    print("Final training value for Precision:", ("%.5f" % P ))
                elif metrics[j] == "recall":
                    print("Final training value for Recall:", ("%.5f" % R ))                    
                elif metrics[j] == "confusion matrix":
                    print("\t\tFinal training value for Confusion Matrix : \n", conf_mat)
        print("\n___________________________________________________\n")
        
        return metrics_output


import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation, PillowWriter  
import numpy as np


class draw:
    def __init__(self, epochs,file_name):
          self.fig = plt.figure()          
          self.ax1 = self.fig.add_subplot()
          self.epoch = epochs
          self.max = None
          self.min = None
          self.file = file_name
          if (self.file == "animation_loss.txt"):
              plt.title("Loss Animation")
              plt.xlabel('Epochs')
              plt.ylabel('Loss')
          elif (self.file == "animation_acc.txt"):
              plt.title("Accuracy Animation")
              plt.xlabel('Epochs')
              plt.ylabel('Accuracy')
          
    def terminate(self):
        open(self.file, "w").close()
        plt.close(self.fig)

        
    def save_gif(self):
          anim = FuncAnimation(self.fig, self.animate , interval=10000, frames  = self.epoch , repeat = True, blit=True)
          writer = PillowWriter(fps=50)  
          gif_name = self.file.replace(".txt","")+ ".gif"
          anim.save(gif_name, writer=writer)
          
    def read (self):
          
        graph_data = open(self.file,'r').read()
        lines = graph_data.split('\n')
        xs = []
        ys = []
        for line in lines:
            if len(line) > 1:
                x, y = line.split(',')
                xs.append(float(x))
                ys.append(float(y))
        self.max = np.max(ys)
        self.min = np.min(ys)
       
        return xs,ys
        
    def animate(self,i):
        
        if (self.file == "animation_loss.txt"):
            self.ax1.set(xlim=(0,self.epoch), ylim=(0,self.max))
            
        elif (self.file == "animation_acc.txt"):
            self.ax1.set(xlim=(0,self.epoch), ylim=(self.min,100))
        
        line_1, = self.ax1.plot([], []) 
        xs , ys = self.read()
    
        line_1.set_data(xs[:i],ys[:i])
        line1 = []
        return line1


 




import numpy as np

# def log(X): #replace log(0)=-9999999 instead of -inf
#     with np.errstate(divide='ignore'):
#         res = np.log(X)
#     res[np.isneginf(res)]=-9999999
#     return res

def compute_multiclass_loss(Y, Y_hat):
    L_sum = np.sum(np.multiply(Y, np.log(Y_hat + 1e-9)))
    m = Y.shape[1]
    L = -(1/m) * L_sum

    return L

def compute_mse_loss(Y, Y_hat):
    m = Y_hat.shape[1]              
    cost= (1 / m) * np.sum(np.sqrt(np.square(np.subtract(Y,Y_hat))))
    
    return cost


import pickle

class saver():
     
    def save(data):
        # Store data (serialize)
        with open('model.pickle', 'wb') as handle:
            pickle.dump(data, handle, protocol=pickle.HIGHEST_PROTOCOL)
        print("file saved")
            
    def restore():   
        # Load data (deserialize)
        with open('model.pickle', 'rb') as handle:
            unserialized_data = pickle.load(handle)
            return unserialized_data





import requests
import numpy as np
import pandas as pd
import os
import pickle
import gzip
import struct
import tarfile


def preprocessing_online(datasetName):
    if datasetName == 'MNIST':
        MNIST_trainFeatures = 'http://yann.lecun.com/exdb/mnist/train-images-idx3-ubyte.gz'
        MNIST_trainLabels = 'http://yann.lecun.com/exdb/mnist/train-labels-idx1-ubyte.gz'
        MNIST_testFeatures = 'http://yann.lecun.com/exdb/mnist/t10k-images-idx3-ubyte.gz'
        MNIST_testLabels = 'http://yann.lecun.com/exdb/mnist/t10k-labels-idx1-ubyte.gz'
        
        R1 = requests.get(MNIST_trainFeatures, allow_redirects=True)
        R2 = requests.get(MNIST_trainLabels, allow_redirects=True)
        R3 = requests.get(MNIST_testFeatures, allow_redirects=True)
        R4 = requests.get(MNIST_testLabels, allow_redirects=True)
        
        open('trainFeaturesMNIST.gz', 'wb').write(R1.content)
        open('trainLabelsMNIST.gz', 'wb').write(R2.content)
        open('testFeaturesMNIST.gz', 'wb').write(R3.content)
        open('testLabelsMNIST.gz', 'wb').write(R4.content)
        
            
        filenames = os.listdir(os.getcwd())
        for file in filenames:
            if 'MNIST.gz' in file:
                if file == 'trainFeaturesMNIST.gz':
                    with gzip.open('trainFeaturesMNIST.gz', 'rb') as f_in:
                        zero, data_type, dims = struct.unpack('>HBB', f_in.read(4))
                        shape = tuple(struct.unpack('>I', f_in.read(4))[0] for d in range(dims))
                        trainFeatures = np.frombuffer(f_in.read(), dtype=np.uint8).reshape(shape)
                if file == 'trainLabelsMNIST.gz':
                    with gzip.open('trainLabelsMNIST.gz', 'rb') as f_in:
                        zero, data_type, dims = struct.unpack('>HBB', f_in.read(4))
                        shape = tuple(struct.unpack('>I', f_in.read(4))[0] for d in range(dims))
                        trainLabels = np.frombuffer(f_in.read(), dtype=np.uint8).reshape(shape)
                if file == 'testFeaturesMNIST.gz':
                    with gzip.open('testFeaturesMNIST.gz', 'rb') as f_in:
                        zero, data_type, dims = struct.unpack('>HBB', f_in.read(4))
                        shape = tuple(struct.unpack('>I', f_in.read(4))[0] for d in range(dims))
                        testFeatures = np.frombuffer(f_in.read(), dtype=np.uint8).reshape(shape)
                if file == 'testLabelsMNIST.gz':
                    with gzip.open('testLabelsMNIST.gz', 'rb') as f_in:
                        zero, data_type, dims = struct.unpack('>HBB', f_in.read(4))
                        shape = tuple(struct.unpack('>I', f_in.read(4))[0] for d in range(dims))
                        testLabels = np.frombuffer(f_in.read(), dtype=np.uint8).reshape(shape)
        
        testLabels = np.atleast_2d(testLabels)
        trainLabels= np.atleast_2d(trainLabels.T)
        testFeatures = testFeatures.reshape(-1,784)
        trainFeatures = trainFeatures.reshape(-1,784)
        trainFeatures = trainFeatures/255
        testFeatures = testFeatures/255 
        idx = np.random.permutation(len(trainFeatures))
        trainFeatures = trainFeatures[idx]
        TrainLabels = trainLabels.T[idx]
        trainLabels = TrainLabels.T
        Dataset = 'MNIST'
    
    elif datasetName == 'CIFAR10':
        CIFAR10 = 'https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz'
        R = requests.get(CIFAR10, allow_redirects=True)
        open('cifar-10-python.tar.gz', 'wb').write(R.content)
        
        trainLabels = []
        trainFeatures = []
        Dataset = 'CIFAR10'
        tarfile.open('cifar-10-python.tar.gz', 'r:gz').extractall(os.getcwd())    
        path = os.getcwd()+'\cifar-10-batches-py'
        filenames = os.listdir(path)
        
        for file in filenames:
            if '_batch' in file:
                with open(path+'\\'+file, 'rb') as fo:
                    dict = pickle.load(fo, encoding='bytes')
                    if 'test' in file:
                        testLabels = dict[b'labels']
                        testFeatures = dict[b'data']
                    else:
                        trainLabels.append(dict[b'labels'])
                        trainFeatures.append(dict[b'data'])
            
        trainLabels = np.array(trainLabels)
        trainFeatures = np.array(trainFeatures)
        trainLabels = trainLabels.reshape((-1,1))
        trainFeatures = trainFeatures.reshape((-1,3072))
        testLabels = np.array(testLabels)
        testLabels =np.atleast_2d(testLabels)
        trainLabels =np.atleast_2d(trainLabels.T)
        trainFeatures = trainFeatures/255
        testFeatures = testFeatures/255 
        idx = np.random.permutation(len(trainFeatures))
        trainFeatures = trainFeatures[idx]
        TrainLabels = trainLabels.T[idx]
        trainLabels = TrainLabels.T
    
    return trainFeatures, trainLabels, testFeatures, testLabels, Dataset



def preprocessing(path):

    #path = r'D:\Gam3a\Neural\Project\Datasets\CIFAR-10\cifar-10-batches-py'
    data_directory_path = path
    filenames = os.listdir(data_directory_path)

    #For CIFAR Dataset:

    if 'csv' not in filenames[0]:

        trainLabels = []
        trainFeatures = []
        Dataset = 'CIFAR10'
        for file in filenames:
            if '_batch' in file:
                with open(path+'/'+file, 'rb') as fo:
                    dict = pickle.load(fo, encoding='bytes')
                    if 'test' in file:
                        testLabels = dict[b'labels']
                        testFeatures = dict[b'data']
                    else:
                        trainLabels.append(dict[b'labels'])
                        trainFeatures.append(dict[b'data'])

        trainLabels = np.array(trainLabels)
        trainFeatures = np.array(trainFeatures)
        trainLabels = trainLabels.reshape((-1,1))
        trainFeatures = trainFeatures.reshape((-1,3072))
        testLabels = np.array(testLabels)
        testLabels =np.atleast_2d(testLabels)
        trainLabels =np.atleast_2d(trainLabels.T)
        trainFeatures = trainFeatures/255
        testFeatures = testFeatures/255 
        idx = np.random.permutation(len(trainFeatures))
        trainFeatures = trainFeatures[idx]
        TrainLabels = trainLabels.T[idx]
        trainLabels = TrainLabels.T
        

    #For MNIST Dataset:

    else:
        trainData = pd.read_csv(path+'/train.csv')
        testData = pd.read_csv(path+'/test.csv')
        trainData = np.array(trainData)
        testData = np.array(testData)
        trainLabels = np.array(trainData[:,0])
        trainFeatures = np.array(trainData[:,1:])
        testFeatures = np.array(testData[:,1:])
        testLabels = np.array(testData[:,0])
        testLabels =np.atleast_2d(testLabels)
        trainLabels =np.atleast_2d(trainLabels.T)
        trainFeatures = trainFeatures/255
        testFeatures = testFeatures/255 
        idx = np.random.permutation(len(trainFeatures))
        trainFeatures = trainFeatures[idx]
        TrainLabels = trainLabels.T[idx]
        trainLabels = TrainLabels.T
        Dataset = 'MNIST'
        
    return trainFeatures, trainLabels, testFeatures, testLabels, Dataset



def labels_to_onehot(labels):
    b = np.zeros((labels.size, labels.max()+1))
    b[np.arange(labels.size),labels] = 1
    b = b.T
    
    return b



#'D:\Gam3a\Neural\Project\Datasets\CIFAR-10\cifar-10-batches-py'



# -*- coding: utf-8 -*-
"""
Created on Sun Jan 24 19:40:55 2021

@author: My Lap
"""
from Terminatetensorflow.Preprocessing import *
from Terminatetensorflow.actvfn import *
from Terminatetensorflow.dense import *
from Terminatetensorflow.LeNet import *
from Terminatetensorflow.LiveAnimation import *
from Terminatetensorflow.loss import *
from Terminatetensorflow.metrics import *
from Terminatetensorflow.model import *
from Terminatetensorflow.pool import *
from Terminatetensorflow.saver import *
from Terminatetensorflow.viz import *
from Terminatetensorflow.conv import *
